/********************************************************************************
					Button Pad Controller USB 
					  Sparfun Electronics
					 Written by Ryan Owens

				       Released under CC BY-SA 3.0 
			     (http://creativecommons.org/licenses/by-sa/3.0/)						    
*********************************************************************************
*/

//***************************************
//			Include Libraries
//***************************************
#include <inttypes.h>
#include <ctype.h>
#include <avr/io.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "octoplododon_master.h"


//***************************************
//			Global Variables
//***************************************
volatile char command_char=0, command_complete=0;
volatile int command_index=0;
volatile char command_buffer[20];
Packet current_command;

unsigned char redf[10*NUMBER_OF_LEDS];
unsigned char bluef[10*NUMBER_OF_LEDS];
unsigned char greenf[10*NUMBER_OF_LEDS];
unsigned char buttons[10*NUMBER_OF_LEDS];

unsigned char command_mode=0;
unsigned char NUMBER_OF_BOARDS=1;
unsigned char system_parameter_id=0;
unsigned char system_parameter_data=0;

ISR (SIG_OUTPUT_COMPARE1B)
{
	cli();
	command_char=getchar245();
	if(command_char=='\r'){
		command_complete=1;
		command_buffer[command_index++]='\n';
	}
	else if(isalnum(command_char) || command_char=='$'){
		command_buffer[command_index++]=command_char;
		command_complete=0;
	}	
	sei();
}

int main (void)
{
    ioinit(); //Boot up defaults
	//printf245("\rBoards: %d", NUMBER_OF_BOARDS);
	
	//Setup default Color Frames
	fillFrame(redf, 0x00);
	fillFrame(bluef, 0x00);
	fillFrame(greenf, 0x00);	
	delay_ms(1000);
	
	sei();
	
	while(1){
	
		//Set CS to indicate Start of Frame
		while(!command_complete);
		parseCommand();
		executeCommand();
		command_complete=0;
			
		if(command_mode==0){
			cli();
			for(int frame_num=0; frame_num <NUMBER_OF_BOARDS; frame_num++){
				sbi(PORTB, SPI_CS);	
				delay_micro(HOLD_TIME);		
			
				//Send RED Frame
				for(int LED=0; LED <NUMBER_OF_LEDS; LED++){
						
					for(int bit_count=0; bit_count <COLOR_BITS; bit_count++){								
						delay_micro(HOLD_TIME);
						cbi(PORTB, SPI_SCLK);					
						
						if((redf[LED+(frame_num*NUMBER_OF_LEDS)]>>bit_count)&0x01)sbi(PORTB, SPI_OUT);
						else cbi(PORTB, SPI_OUT);
						
						delay_micro(HOLD_TIME);
						sbi(PORTB, SPI_SCLK);									
					}
				}
				
				//Send Green Frame
				for(int LED=0; LED <NUMBER_OF_LEDS; LED++){
						
					for(int bit_count=0; bit_count <COLOR_BITS; bit_count++){
						delay_micro(HOLD_TIME);
						cbi(PORTB, SPI_SCLK);					
						
						if((greenf[LED+(frame_num*NUMBER_OF_LEDS)]>>bit_count)&0x01)sbi(PORTB, SPI_OUT);
						else cbi(PORTB, SPI_OUT);
						
						delay_micro(HOLD_TIME);
						sbi(PORTB, SPI_SCLK);
					}
				}
				
				//Send Blue Frame
				for(int LED=0; LED <NUMBER_OF_LEDS; LED++){
						
					for(int bit_count=0; bit_count <COLOR_BITS; bit_count++){	
						delay_micro(HOLD_TIME);
						cbi(PORTB, SPI_SCLK);					
						
						if((bluef[LED+(frame_num*NUMBER_OF_LEDS)]>>bit_count)&0x01)sbi(PORTB, SPI_OUT);
						else cbi(PORTB, SPI_OUT);
						
						delay_micro(HOLD_TIME);	
						sbi(PORTB, SPI_SCLK);															
					}
				}

				//Receive the button data
				for(int LED=0; LED<NUMBER_OF_LEDS; LED++){
					for(int bit_count=0; bit_count<COLOR_BITS; bit_count++){	
						delay_micro(HOLD_TIME);
						cbi(PORTB, SPI_SCLK);
						delay_micro(HOLD_TIME);
						sbi(PORTB, SPI_SCLK);
						
						if(PINB & (1<<SPI_IN))buttons[LED+frame_num*NUMBER_OF_LEDS]|=(1<<bit_count);
						else buttons[LED+frame_num*NUMBER_OF_LEDS]&=~(1<<bit_count);		

					}
				}
				//Clear CS to indicate End of Frame
				cbi(PORTB, SPI_CS);
				
				//Allow Boards to handle color data
				delay_micro(HOLD_TIME*40);				
				
				
			}	//Done sending the entire frame
			sendButtonValues();
			sei();
			
		}
		else{
			command_mode=0;	//Reset the command mode flag
			
			delay_micro(HOLD_TIME);
			//Send the signal sequence to SPI board telling it to enter command mode
			cbi(PORTB, SPI_CS);
			cbi(PORTB, SPI_SCLK);
			cbi(PORTB, SPI_OUT);
			delay_micro(HOLD_TIME);
			
			sbi(PORTB, SPI_OUT);
			delay_micro(HOLD_TIME);
			
			sbi(PORTB, SPI_SCLK);
			delay_micro(HOLD_TIME);
			cbi(PORTB, SPI_SCLK);
			delay_micro(HOLD_TIME);
			
			cbi(PORTB, SPI_OUT);
			delay_micro(HOLD_TIME);
			
			//Send the Command ID byte
			for(int id_bit=0; id_bit<8; id_bit++){
				delay_micro(HOLD_TIME);
				cbi(PORTB, SPI_SCLK);					
				
				if((system_parameter_id>>id_bit)&0x01)sbi(PORTB, SPI_OUT);
				else cbi(PORTB, SPI_OUT);
				
				delay_micro(HOLD_TIME);	
				sbi(PORTB, SPI_SCLK);			
			}
			//Send the Command Parameter Byte
			for(int data_bit=0; data_bit<8; data_bit++){
				delay_micro(HOLD_TIME);
				cbi(PORTB, SPI_SCLK);					
				
				if((system_parameter_data>>data_bit)&0x01)sbi(PORTB, SPI_OUT);
				else cbi(PORTB, SPI_OUT);
				
				delay_micro(HOLD_TIME);	
				sbi(PORTB, SPI_SCLK);			
			}
			//End the Command Mode
			delay_micro(HOLD_TIME);
			cbi(PORTB, SPI_SCLK);
			delay_micro(HOLD_TIME);
			
		}
	}

	return (0);
}

//Fills a color frame with the designated color
void fillFrame(volatile uint8_t* src, uint8_t fill)
{
    int i;
    for (i = 0; i < NUMBER_OF_BOARDS*NUMBER_OF_LEDS; i++)
    {
        src[i] = fill;
    }
}

//Initializes the I/O pins, the interrupts and the system parameters
void ioinit (void)
{

    //1 = output, 0 = input
	DDRB |= (1<<SPI_OUT) | (1<<SPI_SCLK) | (1<<SPI_CS);	//Port B OUtputs
	DDRB &= ~(1<<SPI_IN); //Port B Inputs
	PORTB |= (1<<SPI_IN);
	cbi(PORTB, SPI_CS);
	sbi(PORTB, SPI_SCLK);
					
    //1 = output, 0 = input   
	DDRC |= WR | RD;
	DDRC &= ~(RXF | TXE | PWREN);
	PORTC |= (RXF | TXE | PWREN);
	
	DDRD &= ~(D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7);
	PORTD |= (D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7);
	
    TCCR1B=(1<<WGM12)|(1<<CS12);	//Divide Clock by 256, Clear Timer on Compare	
	OCR1AH=(0x02);				//Set Compare Value to 0x260 for ~1600Hz Interrupts
	OCR1AL=(0x60);	
	TIMSK1 = (1<<OCIE1B);			//Enable Interrupts on Timer 1 Compare
	
	//Load the number of boards from EEPROM
	NUMBER_OF_BOARDS = read_from_EEPROM(NUMBER_OF_BOARDS_ADDRESS);
	if(NUMBER_OF_BOARDS == 0 || NUMBER_OF_BOARDS > 16){
		NUMBER_OF_BOARDS=1;
		write_to_EEPROM(NUMBER_OF_BOARDS_ADDRESS, NUMBER_OF_BOARDS);
	}	
	
	delay_ms(500);
	sei();
}   

//General short delays
void delay_ms(uint16_t x)
{
  uint8_t y, z;
  for ( ; x > 0 ; x--){
    for ( y = 0;  y < 40 ; y++){
      for ( z = 0 ; z < 40 ; z++){
        asm volatile ("nop");
      }
    }
  }
}

//General short delays
void delay_micro(uint16_t x)
{
  uint8_t y;
  for ( ; x > 0 ; x--){
    for ( y = 0;  y < 4 ; y++){
		asm volatile ("nop");
    }
  }
}

//Generatres a "random" new color for the input color arrays
void newcolor(uint8_t * r, uint8_t * g, uint8_t * b)
{

  uint16_t k = *r | (*g << 3) | (*b << 6);
  uint8_t i,j;
  

  i = k&0xff;
  j = k >> 8;
  while((i*j)==0)
  {
    i=i+j+1;
    j=j+3;
  }
  k=i*j;
  
  *r=k&7;
  *g=(k>>3)&7;
  *b=(k>>6)&7;
}

//Configures the I/O pins to output to the FT245
void setDataOutputs(void){
    //1 = output, 0 = input
	PORTD &= ~(D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7); //Disable Pullups (Intermediate Step)
	DDRD |= (D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7);
}

//Configures the I/O pins to input from the FT245
void setDataInputs(void){
    //1 = output, 0 = input	
	DDRD &= ~(D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7);
	PORTD |= (D0 | D1 | D2 | D3 | D4 | D5 | D6 | D7); //Enable Pullups
}

//Printf a string to the FT245. Works like printf
void printf245(const char *text, ...){
	char character=0, temp=0;
	char decimal[6];
	int index=0;
	
	va_list next_argument;		//Create a pointer for the variable argument
	va_start(next_argument, text);	//Initialize the pointer to the next argument in the list
	
	setDataOutputs();
	character=text[index];
	while(character!='\0'){
		if(USBwriteAllowed()){
			if(character=='%'){	//Need to access an argument value!
				character = (char)va_arg(next_argument, int);	//Get the value of the current argument
				
				if(text[index+1]=='d'){
					itoa2(character, decimal);
					for(int i=0; decimal[i]!='\0'; i++)printf245("%c", decimal[i]);
					index+=1;
				}
				else if(text[index+1]=='x'){
					temp=character;
					temp=temp>>4;
					if(temp>9)temp+=('A'-10);
					else temp+='0';
					printf245("%c", temp);
					character=character&0x0f;
					if(character>9)character+=('A'-10);
					else character+='0';
					printf245("%c", character);
					index+=1;
				}
				else if(text[index+1]=='c'){
					PORTD = character;				
					clearWR();
					delay_ms(1);
					setWR();
					index+=1;
				}				
			}
			else{
				PORTD = character;				
				clearWR();
				delay_ms(1);
				setWR();
			}
			character=text[++index];
		}
	}
	va_end(next_argument);	//Relinquish the pointer to system memory
}

//Returs a character from the FT245
char getchar245(void){
	char character=0;
	
	setDataInputs();
	if(USBreadAllowed()){		//If RX Flag is low, lets get the data from the FIFO
		clearRD();			//Clock the next FIFO data onto the D0-D7 lines
		delay_micro(HOLD_TIME);
		character = PIND;		//Get data from FT245		
		setRD();			//Relinquish control of the Data lines
	}
	
	return character;
}

/* reverse:  reverse string s in place */
void reverse(char s[])
{
    int c, i, j;

    for (i = 0, j = strlen(s)-1; i<j; i++, j--) {
        c = s[i];
        s[i] = s[j];
        s[j] = c;
    }
}

/* itoa:  convert n to characters in s */
void itoa2(int n, char s[])
{
    int i=0;

    do {       /* generate digits in reverse order */
        s[i++] = n % 10 + '0';   /* get next digit */
    } while ((n /= 10) > 0);     /* delete it */

    s[i] = '\0';
    reverse(s);
}

//Parses an input command to discover the ID and parameters
void parseCommand(void){

	cli();	//Pause interrupts so we don't receive any characters while we parse the current command
	//Convert the command (which has been input in ascii format) to a hex string, but keep the '$'
	ascii2hex(command_buffer);

	//Check to see if we received the start character.  If not, we didn't receive a command!
	if(command_buffer[0]=='$'){
		current_command.ID=command_buffer[1];
		if(current_command.ID==SET_PIXEL_COMMAND){
				//Get the command parameters from the UART buffer
				for(int current_parameter=0; current_parameter<PIXEL_PAYLOAD_LENGTH; current_parameter++){
					current_command.Parameters[current_parameter]=command_buffer[current_parameter+2];
				}
		}
		
		if(current_command.ID==SET_BOARD_COMMAND){	
				//Get the command parameters from the UART buffer
				for(int current_parameter=0; current_parameter<BOARD_PAYLOAD_LENGTH; current_parameter++){
					current_command.Parameters[current_parameter]=command_buffer[current_parameter+2];
				}
		}		

		if(current_command.ID==CHANGE_NUM_BOARDS){	
				//Get the command parameters from the UART buffer
				for(int current_parameter=0; current_parameter<NUM_BOARDS_PAYLOAD_LENGTH; current_parameter++){
					current_command.Parameters[current_parameter]=command_buffer[current_parameter+2];
				}
		}		
	}
	for(int i=0; i<20; i++)command_buffer[i]='\0';	//Clear the current command out of the buffer
	command_index=0;
	sei();
}

//Use to find a string within a search string
//search : "hithere\0"
//find : "the\0" - return OK
//***************************** UNTESTED *********************************
char string_compare(const char *search_string, const char *find_string)
{

    int find_spot, search_spot;
    char spot_character, search_character;
    find_spot = 0;

    for(search_spot = 0 ; ; search_spot++)
    {
        if(find_string[find_spot] == '\0') return OK; //We've reached the end of the search string - that's good!
        if(search_string[search_spot] == '\0') return ERROR; //End of string found

        spot_character = find_string[find_spot]; //Compiler limit
        search_character = search_string[search_spot]; //Compiler limit

        if(spot_character == search_character) //We found another character
            find_spot++; //Look for the next spot in the search string
        else if(find_spot > 0) //No character found, so reset the find_spot
            find_spot = 0;
    }

    return 0;
}


void executeCommand(void){
	if(current_command.ID==SET_PIXEL_COMMAND){
		setPixel(current_command.Parameters[PIXEL_BOARD_PARAMETER], current_command.Parameters[PIXEL_OFFSET_PARAMETER], current_command.Parameters[PIXEL_RED_VAL],
			current_command.Parameters[PIXEL_GREEN_VAL], current_command.Parameters[PIXEL_BLUE_VAL]);
		command_mode=0;
	}
	
	if(current_command.ID==SET_BOARD_COMMAND){
		setBoard(current_command.Parameters[BOARD_RED], current_command.Parameters[BOARD_GREEN], current_command.Parameters[BOARD_BLUE]);
		command_mode=0;
	}
	
	if(current_command.ID==GET_BUTTONS){
		//Don't change the frame.  Just send the current frame to get the button array from the slaves
		command_mode=0;
	}
	
	if(current_command.ID==CHANGE_NUM_BOARDS){
		NUMBER_OF_BOARDS=current_command.Parameters[0];
		write_to_EEPROM(NUMBER_OF_BOARDS_ADDRESS, NUMBER_OF_BOARDS);
		
		command_mode=1;
		system_parameter_id=COMMAND_NUM_BOARDS;
		system_parameter_data=NUMBER_OF_BOARDS;
	}
	else command_mode=0;
}

void ascii2hex(volatile char *ascii_string)
{
	char hex_value;
	char hex_string[20];
	int hex_size=0;
	int ascii_index=1;
	
	//Convert each character in the received command buffer to a hex value
	for(int string_index=1; ascii_string[string_index] != '\n'; string_index++){
		hex_value=ascii_string[string_index];
		if(isdigit(hex_value))hex_value-='0';
		else if(isupper(hex_value)){
			hex_value-='A';
			hex_value+=10;
		}
		else if(islower(hex_value)){
			hex_value-='a';
			hex_value+=10;
		}
		hex_string[string_index-1]=hex_value;
		hex_size++;
	}
	hex_string[hex_size]='\n';

	//Now that we have the hex values, we need to concatonate each "pair" since we received the command serially.
	//(i.e. a received command: "01A23E" is 6 characters, but the hex command is only 3 characters)
	for(int index=0; hex_string[index]!='\n'; index++){
		
		//Combine two values in the hex string to form a single hex byte
		hex_value=((hex_string[index]&0x0f)<<4) | (hex_string[index+1]&0x0f);
		
		//Put the new hex value back in the ascii string for use by the main program
		ascii_string[ascii_index++]=hex_value;
		
		//Increment the hex index since we are only getting "pairs"
		index+=1;
		hex_size=index;
	}
	ascii_index=1;
	ascii_string[hex_size/2+2]='\n';
	
}

void setPixel(int board, int offset, char r, char g, char b){
	int position=(board*NUMBER_OF_LEDS)+offset;

	redf[position]=r;
	greenf[position]=g;
	bluef[position]=b;
}

void setBoard(char r, char g, char b){
	fillFrame(redf, r);
	fillFrame(greenf, g);
	fillFrame(bluef, b);	
}

void sendButtonValues(void){
	unsigned char pressed_button_index[NUMBER_OF_BOARDS*4];
	int number_of_buttons_pressed=0;	
	int current_index=0;
	
	for(int i=0; i<NUMBER_OF_LEDS*NUMBER_OF_BOARDS;i++){
		if(buttons[i]==PRESSED){
			pressed_button_index[current_index++]=i;
			number_of_buttons_pressed+=1;
		}
	}
	
	int button_coordinates[number_of_buttons_pressed*2];
	int board, board_offset, new_index;
	
	for(int i=0; i<number_of_buttons_pressed; i++){
		board=pressed_button_index[i]/NUMBER_OF_LEDS;
		board_offset=board*NUMBER_OF_LEDS;
		
		new_index=pressed_button_index[i]-board_offset;

		button_coordinates[i*2]=board;
		button_coordinates[(i*2)+1]=new_index;
	}	
	printf245("$%x", number_of_buttons_pressed);
	for(int i=0; i<number_of_buttons_pressed*2; i++) printf245("%x", button_coordinates[i]);
	printf245("\n\r");
	
}

//Description: Writes an unsigned char(Data)  to the EEPROM at the given Address
//Pre: Unsigned Int Address contains address to be written to
//	 Unsigned Char Data contains data to be written
//Post: EEPROM "Address" contains the "Data"
//Usage: write_to_EEPROM(0, 'A');
void write_to_EEPROM(unsigned int Address, unsigned char Data)
{
    //Interrupts are globally disabled!
	
	while(EECR & (1<<EEPE)); //Wait for last Write to complete
	//May need to wait for Flash to complete also!
	EEAR = Address;			//Assign the Address Register with "Address"
	EEDR=Data;				//Put "Data" in the Data Register
	EECR |= (1<<EEMPE); 	//Write to Master Write Enable
	EECR |= (1<<EEPE);  	//Start Write by setting EE Write Enable
	
}

//Description: Reads the EEPROM data at "Address" and returns the character
//Pre: Unsigned Int Address is the address to be read
//Post: Character at "Address" is returned
//Usage: 	unsigned char Data;
//		Data=read_from_EEPROM(0);
unsigned char read_from_EEPROM(unsigned int Address)
{
	//Interrupts are globally disabled!
	
	while(EECR & (1<<EEPE));	//Wait for last Write to complete
	EEAR = Address;				//Assign the Address Register with "Address"
	EECR |= (1<<EERE); 			//Start Read by writing to EER
	return EEDR;				//EEPROM Data is returned
}
