//*******************************************************
//
//		Button Pad Controller USB Definitions
//				SparkFun Electronics
//   		   Written By Ryan Owens
//			
//*******************************************************
//HARDWARE NOTE: High Fuse should be set to 0xDF
//				 Low Fuse should be set to 0xFF

//*******************************************************
//					General Definitions
//*******************************************************
#define COLOR_BITS 	8
#define HOLD_TIME	20

#define OK	1
#define ERROR	0
//#define NUMBER_OF_BOARDS	1
#define NUMBER_OF_LEDS		16
#define NUMBER_OF_BOARDS_ADDRESS	0

//*******************************************************
//				Parameter Command Definitions
//*******************************************************
//Command IDs
#define SET_PIXEL_COMMAND	0
#define SET_BOARD_COMMAND	1
#define GET_BUTTONS			2
#define CHANGE_NUM_BOARDS	3

//Command Payload Lengths
#define PIXEL_PAYLOAD_LENGTH		5
#define BOARD_PAYLOAD_LENGTH		3
#define BUTTONS_PAYLOAD_LENGTH		0
#define NUM_BOARDS_PAYLOAD_LENGTH	1

//Pixel Payload Parameters
#define PIXEL_BOARD_PARAMETER	0
#define PIXEL_OFFSET_PARAMETER	1
#define PIXEL_RED_VAL		2
#define PIXEL_GREEN_VAL		3
#define PIXEL_BLUE_VAL		4

//Board Payload Parameters
#define BOARD_RED	0
#define BOARD_GREEN	1
#define BOARD_BLUE	2

//Button State Definitions
#define NOT_PRESSED	0xFF
#define PRESSED		0x00

//*******************************************************
//				Command Mode Byte Definitions
//*******************************************************
#define COMMAND_NUM_BOARDS	1

//*******************************************************
//					ATmega168 Pin Definitions
//*******************************************************
//SPI Communication Pins
#define SPI_SCLK	5
#define SPI_IN		4
#define SPI_OUT		3
#define SPI_CS		2

//FT245 Control Pin Definitions
#define PWREN	(1<<4) //PC4 (Input to AVR)
#define RD		(1<<3) //PC3 (Output from AVR)
#define WR		(1<<2) //PC2 (Output from AVR)
#define TXE		(1<<1) //PC1 (Input to AVR)
#define RXF		(1<<0) //PC0 (Input to AVR)

//FT245 Data Pins (These pins switch between being I/O)
#define D0		(1<<0) //PD0
#define D1		(1<<1) //PD1
#define D2		(1<<2) //PD2
#define D3		(1<<3) //PD3
#define D4		(1<<4) //PD4
#define D5		(1<<5) //PD5
#define D6		(1<<6) //PD6
#define D7		(1<<7) //PD7

//*******************************************************
//					Global Macros
//*******************************************************
#define sbi(var, mask)   ((var) |= (uint8_t)(1 << mask))
#define cbi(var, mask)   ((var) &= (uint8_t)~(1 << mask))

//Macros for interfacing with the FT245
#define USBreadAllowed()	(PINC & RXF) != RXF
#define USBwriteAllowed()  	(PINC & TXE) != TXE
#define USBConfigured() 	(PINC & PWREN) != PWREN
#define setRD()		PORTC |= RD;
#define clearRD()	PORTC &= ~RD;
#define setWR()		PORTC |= WR;
#define clearWR()	PORTC &= ~WR;

//*******************************************************
//					Global Functions
//*******************************************************
void delay_ms(uint16_t x); // general purpose delay
void delay_micro(uint16_t x); // general purpose delay
void fill16(volatile uint8_t* src, uint8_t fill);
void fillFrame(volatile uint8_t* src, uint8_t fill);
void newcolor(uint8_t * r, uint8_t * g, uint8_t * b);
void clock_byte(volatile uint16_t my_byte, volatile uint8_t rowNum);
void ioinit (void);      // initializes IO
void setDataOutputs(void);
void setDataInputs(void);
void printf245(const char *text, ...);
char getchar245(void);
void reverse(char s[]);
void itoa2(int n, char s[]);
void parseCommand(void);
void executeCommand(void);
char string_compare(const char *search_string, const char *find_string);
void ascii2hex(volatile char *ascii_string);
void setPixel(int board, int offset, char r, char g, char b);
void setBoard(char r, char g, char b);
void sendButtonValues(void);
unsigned char read_from_EEPROM(unsigned int Address);
void write_to_EEPROM(unsigned int Address, unsigned char Data);

//*******************************************************
//					Structure Definitions
//*******************************************************
typedef struct{
	char ID;
	char Parameters[14];
	//char Checksum;
} Packet;

